import logging
import os
import pickle

import numpy as np
import torch
from tqdm import tqdm

from ..util.constants import DEFAULT_FEATURES, IS_BALL, IS_BOOST, BALL_COLS, PLAYER_COLS, DEFAULT_FEATURES_STR, IS_BLUE, \
    IS_ORANGE, POS_X, POS_Y, POS_Z, BOOST_AMOUNT, IS_DEMOED
from ..util.util import boost_locations, rotator_to_matrix

X_DATASET_NAMES = ("balls", "boosts", "players")
Y_DATASET_NAMES = ("next_goal", "next_touch", "next_boost", "next_demo",
                   "control_throttle", "control_steer", "control_pitch", "control_yaw", "control_roll",
                   "control_jump", "control_boost", "control_handbrake")


def iterate_replays(bc_api, replay_ids=None, replay_folder=None, cache_folder=None):
    if replay_ids is not None:
        assert replay_folder is not None
        import ballchasing as bc
        bc_api: bc.Api
        for replay_id in tqdm(replay_ids, "Replay download"):
            if not os.path.exists(os.path.join(replay_folder, replay_id + ".replay")):
                bc_api.download_replay(replay_id, replay_folder)

    if replay_folder is not None and cache_folder is not None:
        for replay_file in tqdm(os.listdir(replay_folder), "Replay processing"):
            cache_file = os.path.join(cache_folder, replay_file.replace(".replay", ".pickle"))
            if os.path.exists(cache_file):
                with open(cache_file, "rb") as handle:
                    processed = pickle.load(handle)
            else:
                try:
                    replay_path = os.path.join(replay_folder, replay_file)
                    processed = replay_to_dfs(replay_path)
                    with open(cache_file, "wb") as handle:
                        pickle.dump(processed, handle)
                except Exception as e:
                    # raise e
                    print(e)
                    continue
            yield convert_dfs(processed)
    elif replay_folder is not None and cache_folder is None:
        for replay_file in tqdm(os.listdir(replay_folder), "Replay processing"):
            try:
                replay_path = os.path.join(replay_folder, replay_file)
                processed = replay_to_dfs(replay_path, skip_ties=False, skip_kickoffs=True)
                yield convert_dfs(processed)
            except Exception as e:
                # raise e
                print(e)
                continue
    elif replay_folder is None and cache_folder is not None:
        for cache_file in tqdm(os.listdir(cache_folder), "Replay loading"):
            with open(cache_file, "rb") as handle:
                processed = pickle.load(handle)
            yield convert_dfs(processed)


def replay_to_dfs(replay, skip_ties=True, skip_kickoffs=True):
    import pandas as pd
    import carball as cb

    if isinstance(replay, (str, bytes)):
        from carball.controls.controls import ControlsCreator
        replay = cb.analyze_replay_file(replay, controls=ControlsCreator(), logging_level=logging.CRITICAL)

    blue_team, orange_team = replay.game.teams
    if blue_team.is_orange:
        blue_team, orange_team = orange_team, blue_team

    frames = pd.DataFrame(index=replay.data_frame.index)

    frames[f"ball/pos_x"] = replay.game.ball["pos_x"].fillna(0.)
    frames[f"ball/pos_y"] = replay.game.ball["pos_y"].fillna(0.)
    frames[f"ball/pos_z"] = replay.game.ball["pos_z"].fillna(0.)

    frames[f"ball/vel_x"] = replay.game.ball["vel_x"].fillna(0.) / 10
    frames[f"ball/vel_y"] = replay.game.ball["vel_y"].fillna(0.) / 10
    frames[f"ball/vel_z"] = replay.game.ball["vel_z"].fillna(0.) / 10

    frames[f"ball/ang_vel_x"] = replay.game.ball["ang_vel_x"].fillna(0.) / 1000
    frames[f"ball/ang_vel_y"] = replay.game.ball["ang_vel_y"].fillna(0.) / 1000
    frames[f"ball/ang_vel_z"] = replay.game.ball["ang_vel_z"].fillna(0.) / 1000

    boost_grabs = pd.DataFrame(columns=["frame", "boost_id", "player"])

    player_index = {}
    player_df = pd.DataFrame(columns=["color", "online_id", "name"])
    i = 0
    for color, team in ("blue", blue_team), ("orange", orange_team):
        for player in team.players:
            player_index[player.online_id] = i
            frames[f"{i}/pos_x"] = player.data["pos_x"].fillna(0.)
            frames[f"{i}/pos_y"] = player.data["pos_y"].fillna(0.)
            frames[f"{i}/pos_z"] = player.data["pos_z"].fillna(0.)

            yaw = player.data["rot_y"].fillna(0.)
            pitch = player.data["rot_x"].fillna(0.)
            roll = player.data["rot_z"].fillna(0.)
            forward, up = rotator_to_matrix(yaw, pitch, roll)

            frames[f"{i}/forward_x"] = forward[0]
            frames[f"{i}/forward_y"] = forward[1]
            frames[f"{i}/forward_z"] = forward[2]

            frames[f"{i}/up_x"] = up[0]
            frames[f"{i}/up_y"] = up[1]
            frames[f"{i}/up_z"] = up[2]

            frames[f"{i}/vel_x"] = player.data["vel_x"].fillna(0.) / 10
            frames[f"{i}/vel_y"] = player.data["vel_y"].fillna(0.) / 10
            frames[f"{i}/vel_z"] = player.data["vel_z"].fillna(0.) / 10

            frames[f"{i}/ang_vel_x"] = player.data["ang_vel_x"].fillna(0.) / 1000
            frames[f"{i}/ang_vel_y"] = player.data["ang_vel_y"].fillna(0.) / 1000
            frames[f"{i}/ang_vel_z"] = player.data["ang_vel_z"].fillna(0.) / 1000

            frames[f"{i}/boost_amount"] = player.data["boost"].fillna(0.) / 2.55

            for col in player.controls.columns:
                frames[f"{i}/{col}_control"] = player.controls[col].astype(float)

            boost_pickups = player.data["boost"].diff() > 0
            for frame, row in player.data[boost_pickups].iterrows():
                xyz = (row["pos_x"], row["pos_y"], row["pos_z"])
                # TODO use boost ID instead
                closest_boost = min(range(len(boost_locations)),
                                    key=lambda j: sum((a - b) ** 2 for a, b in zip(xyz, boost_locations[j])))
                boost_grabs.loc[len(boost_grabs)] = [frame, closest_boost, i]

            player_df.loc[len(player_df)] = [color, player.online_id, player.name]
            i += 1

    rallies = pd.DataFrame(columns=["start_frame", "end_frame", "team"])
    for kf1, kf2 in zip(replay.game.kickoff_frames, replay.game.kickoff_frames[1:] + [replay.game.frames.index[-1]]):
        for goal in replay.game.goals:
            if kf1 < goal.frame_number < kf2:
                rallies.loc[len(rallies)] = [kf1, goal.frame_number, ["blue", "orange"][goal.player_team]]
                break
        else:  # No goal between kickoffs
            rallies.loc[len(rallies)] = [kf1, kf2, None]

    demos = pd.DataFrame(columns=["frame", "attacker", "victim"])
    for demo in replay.game.demos:
        frame = demo.get("frame_number", None)
        attacker = demo.get("attacker", None)
        victim = demo.get("victim", None)

        if frame and attacker and victim:
            demos.loc[len(demos)] = [frame, player_index[attacker.online_id], player_index[victim.online_id]]

    touches = pd.DataFrame(columns=["frame", "player"])
    for touch in replay.protobuf_game.game_stats.hits:
        touches.loc[len(touches)] = [touch.frame_number, player_index[touch.player_id.id]]

    if skip_ties:
        goal_rallies = rallies[rallies["team"].notna()]
        frames = frames.loc[
            np.r_[tuple(slice(row["start_frame"], row["end_frame"]) for _, row in goal_rallies.iterrows())]]
    elif skip_kickoffs:
        goal_rallies = rallies
        frames = frames.loc[
            np.r_[tuple(slice(row["start_frame"], row["end_frame"]) for _, row in goal_rallies.iterrows())]]

    # if isinstance(frame_mode, int):
    #     frames = frames.iloc[np.random.randint(0, frame_mode)::frame_mode]
    # else:
    #     frames = frames.sample(frac=frame_mode)

    return {"frames": frames, "rallies": rallies, "touches": touches, "boost_grabs": boost_grabs, "demos": demos,
            "players": player_df}


def convert_dfs(dfs, tensors=False, frame_mode=1):
    frames, rallies, touches, boost_grabs, demos, player_df = \
        (dfs[k] for k in ("frames", "rallies", "touches", "boost_grabs", "demos", "players"))

    control_cols = frames.filter(regex="_control$").columns
    frames[control_cols] = frames[control_cols].shift()
    frames = frames.iloc[np.random.randint(0, frame_mode)::frame_mode]

    if tensors:
        convert_fn = lambda x: torch.from_numpy(x).to(torch.float)
    else:
        convert_fn = lambda x: x

    x, y = get_base_features(len(frames), 6, tensors=tensors)
    x_ball, x_boost, x_players = x
    (y_score, y_next_touch, y_collect, y_demo,
     y_throttle, y_steer, y_pitch, y_yaw, y_roll, y_jump, y_boost, y_handbrake) = y

    x_ball[:, 0, BALL_COLS] = \
        convert_fn(
            frames[[f"ball/{col}" for col in map(DEFAULT_FEATURES_STR.__getitem__, BALL_COLS)]].values)

    bins = np.array([-50, -0.1, 0.1])  # np.arange(0.25 / 2 - 1, 1, 0.25)

    for i, row in player_df.iterrows():
        color = row["color"]
        if color == "blue":
            x_players[:, i, IS_BLUE] = 1
        elif color == "orange":
            x_players[:, i, IS_ORANGE] = 1
        else:
            raise ValueError(color)

        x_players[:, i, PLAYER_COLS[:-3]] = \
            convert_fn(frames[[f"{i}/{col}" for col in
                               map(DEFAULT_FEATURES_STR.__getitem__, PLAYER_COLS[:-3])]].fillna(0.).values)

        throttle = np.digitize(frames[f"{i}/throttle_control"].fillna(-100).values, bins) - 1
        steer = np.digitize(frames[f"{i}/steer_control"].fillna(-100).values, bins) - 1
        pitch = np.digitize(frames[f"{i}/pitch_control"].fillna(-100).values, bins) - 1
        yaw = np.digitize(frames[f"{i}/yaw_control"].fillna(-100).values, bins) - 1
        roll = np.digitize(frames[f"{i}/roll_control"].fillna(-100).values, bins) - 1
        jump = frames[f"{i}/jump_control"].fillna(-100).astype(int).values
        boost = frames[f"{i}/boost_control"].fillna(-100).astype(int).values
        handbrake = frames[f"{i}/handbrake_control"].fillna(-100).astype(int).values

        throttle[throttle < 0] = -100
        steer[steer < 0] = -100
        pitch[pitch < 0] = -100
        yaw[yaw < 0] = -100
        roll[roll < 0] = -100
        jump[jump < 0] = -100
        boost[boost < 0] = -100
        handbrake[handbrake < 0] = -100

        y_throttle[:, i] = convert_fn(throttle)
        y_steer[:, i] = convert_fn(steer)
        y_pitch[:, i] = convert_fn(pitch)
        y_yaw[:, i] = convert_fn(yaw)
        y_roll[:, i] = convert_fn(roll)
        y_jump[:, i] = convert_fn(jump)
        y_boost[:, i] = convert_fn(boost)
        y_handbrake[:, i] = convert_fn(handbrake)

    for i, (x, y, z) in enumerate(boost_locations):
        x_boost[:, i, POS_X] = x
        x_boost[:, i, POS_Y] = y
        x_boost[:, i, POS_Z] = z
        x_boost[:, i, BOOST_AMOUNT] = 100. if z > 72 else 12

    for _, rally in rallies.iterrows():
        indices = (rally["start_frame"] <= frames.index) & (frames.index <= rally["end_frame"])
        team = rally["team"]
        if team == "blue":
            y_score[indices] = 0
        elif team == "orange":
            y_score[indices] = 1
        else:
            # print("Team:", team)
            pass

        rally_touches = touches[
            (rally["start_frame"] <= touches["frame"]) & (touches["frame"] < rally["end_frame"])]
        rally_demos = demos[(rally["start_frame"] <= demos["frame"]) & (demos["frame"] < rally["end_frame"])]
        rally_collects = boost_grabs[
            (rally["start_frame"] <= boost_grabs["frame"]) & (boost_grabs["frame"] < rally["end_frame"])]

        last_frame = rally["start_frame"]
        for _, touch in rally_touches.iterrows():
            frame = touch["frame"]
            player = touch["player"]

            indices = (last_frame <= frames.index) & (frames.index < frame)
            y_next_touch[indices, 0] = player
            last_frame = frame

        for i, row in player_df.iterrows():
            player_demos = rally_demos[rally_demos["attacker"] == i]
            player_collects = rally_collects[rally_collects["player"] == i]

            last_frame = rally["start_frame"]
            for _, boost_grab in player_collects.iterrows():
                frame = boost_grab["frame"]
                boost_id = boost_grab["boost_id"]

                if boost_id in (3, 4, 15, 18, 29, 30):
                    collected_indices = (frames.index - frame >= 0) & (frames.index - frame < 10 * 30)
                else:
                    collected_indices = (frames.index - frame >= 0) & (frames.index - frame < 4 * 30)

                x_boost[collected_indices, boost_id, BOOST_AMOUNT] = 0.
                x_boost[collected_indices, boost_id, IS_DEMOED] = 1.

                indices = (last_frame <= frames.index) & (frames.index < frame)
                y_collect[indices, i] = boost_id
                last_frame = frame

            last_frame = rally["start_frame"]
            for _, demo in player_demos.iterrows():
                frame = demo["frame"]

                j = demo["victim"]

                demoed_indices = (frames.index - frame >= 0) & (frames.index - frame < 30 * 3)
                x_players[demoed_indices, j, IS_DEMOED] = 1

                indices = (last_frame <= frames.index) & (frames.index < frame)
                y_demo[indices, i] = j
                last_frame = frame
    x_data = [x_ball, x_boost, x_players]
    y_data = [y_score, y_next_touch, y_collect, y_demo,
              y_throttle, y_steer, y_pitch, y_yaw, y_roll, y_jump, y_boost, y_handbrake]

    # indices = np.concatenate([y_score[20:] - y_score[:-20] != 0, np.array([False] * 20)])
    # x_data = [v[indices] for v in x_data]
    # y_data = [v[indices] for v in y_data]
    return x_data, y_data


def get_base_features(size, n_players, n_balls=1, n_boosts=34, include_y=True, tensors=False, n_features=None):
    if n_features is None:
        n_features = len(DEFAULT_FEATURES)

    if tensors:
        initializer = torch
    else:
        initializer = np

    x_ball = initializer.zeros((size, n_balls, n_features))
    x_boost = initializer.zeros((size, n_boosts, n_features))
    x_players = initializer.zeros((size, n_players, n_features))

    x_ball[:, :, IS_BALL] = 1
    x_boost[:, :, IS_BOOST] = 1

    if not include_y:
        return [x_ball, x_boost, x_players]

    y_score = initializer.full((size,), -100)
    y_next_touch = initializer.full((size, n_balls,), -100)
    y_collect = initializer.full((size, n_players,), -100)
    y_demo = initializer.full((size, n_players,), -100)

    y_throttle = initializer.full((size, n_players), -100)
    y_steer = initializer.full((size, n_players), -100)
    y_pitch = initializer.full((size, n_players), -100)
    y_yaw = initializer.full((size, n_players), -100)
    y_roll = initializer.full((size, n_players), -100)
    y_jump = initializer.full((size, n_players), -100)
    y_boost = initializer.full((size, n_players), -100)
    y_handbrake = initializer.full((size, n_players), -100)

    return [x_ball, x_boost, x_players], \
           [y_score, y_next_touch, y_collect, y_demo,
            y_throttle, y_steer, y_pitch, y_yaw, y_roll, y_jump, y_boost, y_handbrake]


def normalize(x_data):
    norms = np.array([
        1.,
        1., 1., 1., 1.,
        2300., 2300., 2300.,
        1., 1., 1.,
        1., 1., 1.,
        2300., 2300., 2300.,
        5.5, 5.5, 5.5,
        100., 1., 1., 1.
    ], dtype=np.float32)[:x_data[0].shape[-1]]

    for i in range(len(x_data)):
        x_data[i] /= norms


def swap_teams(x_data, y_data=None, indices=None):
    if indices is None:
        indices = slice(None)

    swaps = np.array([
        1.,
        1., 1., 1., 1.,
        -1., -1., 1.,
        -1., -1., 1.,
        -1., -1., 1.,
        -1., -1., 1.,
        -1., -1., 1.,
        1., 1., 1., 1.
    ])[:x_data[0].shape[-1]]

    x_ball, x_boost, x_players = x_data

    x_ball[indices, :, :] *= swaps

    x_boost[indices, :, :] = x_boost[indices, :, :] * swaps

    x_players[indices, :, :] = x_players[indices, :, :] * swaps
    x_players[indices, :, [IS_BLUE, IS_ORANGE]] = x_players[indices, :,
                                                  [IS_ORANGE, IS_BLUE]]  # Swap blue/orange indicator

    if y_data is None:
        return

    (y_score, y_next_touch, y_collect, y_demo,
     y_throttle, y_steer, y_pitch, y_yaw, y_roll, y_jump, y_boost, y_handbrake) = y_data

    # n_players = x_players.shape[1] - 1

    y_score[indices] = 1 - y_score[indices]
    y_score[y_score > 1] = -100

    # y_next_touch[indices, :] = n_players - y_next_touch[indices, :]
    # y_next_touch[y_next_touch > n_players] = -100


def swap_left_right(x_data, y_data=None, indices=None):
    swaps = np.array([
        1.,
        1., 1., 1., 1.,
        -1., 1., 1.,
        -1., 1., 1.,
        -1., 1., 1.,
        -1., 1., 1.,
        -1., 1., 1.,
        1., 1., 1., 1.
    ])[:x_data[0].shape[-1]]
    if indices is None:
        indices = slice(None)
    # Not really necessary for transformer
    x_ball, x_boost, x_players = x_data

    x_ball[indices, :, :] *= swaps
    x_boost[indices, :, :] *= swaps
    x_players[indices, :, :] *= swaps

    if y_data is None:
        return

    (y_score, y_next_touch, y_collect, y_demo,
     y_throttle, y_steer, y_pitch, y_yaw, y_roll, y_jump, y_boost, y_handbrake) = y_data

    y_steer[indices, :] = 2 - y_steer[indices, :]
    y_steer[y_steer > 2] = -100

    y_yaw[indices, :] = 2 - y_yaw[indices, :]
    y_yaw[y_yaw > 2] = -100

    y_roll[indices, :] = 2 - y_roll[indices, :]
    y_roll[y_roll > 2] = -100



import math
import numpy as np
from typing import Any, List
from rlgym.utils import common_values
from rlgym.utils.gamestates import PlayerData, GameState, PhysicsObject
from rlgym.utils.obs_builders import ObsBuilder


class AdvancedObs(ObsBuilder):
    POS_STD = 2300 # If you read this and wonder why, ping Rangler in the dead of night.
    ANG_STD = math.pi

    def __init__(self):
        super().__init__()

    def reset(self, initial_state: GameState):
        pass

    def build_obs(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> Any:

        if player.team_num == common_values.ORANGE_TEAM:
            inverted = True
            ball = state.inverted_ball
            pads = state.inverted_boost_pads
        else:
            inverted = False
            ball = state.ball
            pads = state.boost_pads

        obs = [ball.position / self.POS_STD,
               ball.linear_velocity / self.POS_STD,
               ball.angular_velocity / self.ANG_STD,
               previous_action,
               pads]

        player_car = self._add_player_to_obs(obs, player, ball, inverted)

        allies = []
        enemies = []

        for other in state.players:
            if other.car_id == player.car_id:
                continue

            if other.team_num == player.team_num:
                team_obs = allies
            else:
                team_obs = enemies

            other_car = self._add_player_to_obs(team_obs, other, ball, inverted)

            # Extra info
            team_obs.extend([
                (other_car.position - player_car.position) / self.POS_STD,
                (other_car.linear_velocity - player_car.linear_velocity) / self.POS_STD
            ])

        obs.extend(allies)
        obs.extend(enemies)
        return np.concatenate(obs)

    def _add_player_to_obs(self, obs: List, player: PlayerData, ball: PhysicsObject, inverted: bool):
        if inverted:
            player_car = player.inverted_car_data
        else:
            player_car = player.car_data

        rel_pos = ball.position - player_car.position
        rel_vel = ball.linear_velocity - player_car.linear_velocity

        obs.extend([
            rel_pos / self.POS_STD,
            rel_vel / self.POS_STD,
            player_car.position / self.POS_STD,
            player_car.forward(),
            player_car.up(),
            player_car.linear_velocity / self.POS_STD,
            player_car.angular_velocity / self.ANG_STD,
            [player.boost_amount,
             int(player.on_ground),
             int(player.has_flip),
             int(player.is_demoed)]])

        return player_car



"""
Data classes to permit the manipulation of environment variables.
"""

from rlgym.utils.gamestates.game_state import GameState
from rlgym.utils.state_setters.wrappers import PhysicsWrapper, CarWrapper
from rlgym.utils.common_values import BLUE_TEAM, ORANGE_TEAM
from typing import List


class StateWrapper(object):
    BLUE_ID1 = 1
    ORANGE_ID1 = 5

    def __init__(self, blue_count: int = 0, orange_count: int = 0, game_state=None):
        """
        StateWrapper constructor. Under most circumstances, users should not expect to instantiate their own StateWrapper objects.

        :param blue_count: Integer indicating the amount of players on the blue team.
        :param orange_count: Integer indicating The amount of players on the orange team.
        :param game_state: GameState object for values to be copied from.

        NOTE: blue_count and orange_count will be ignored if a GameState object is passed.
        """
        if game_state is None:
            self.ball: PhysicsWrapper = PhysicsWrapper()
            self.cars: List[CarWrapper] = []
            for i in range(blue_count):
                self.cars.append(CarWrapper(BLUE_TEAM, StateWrapper.BLUE_ID1 + i))
            for i in range(orange_count):
                self.cars.append(CarWrapper(ORANGE_TEAM, StateWrapper.ORANGE_ID1 + i))
        else:
            self._read_from_gamestate(game_state)

    def _read_from_gamestate(self, game_state: GameState):
        """
        A function to modify the StateWrapper with values read in from a GameState object.
        """
        self.ball: PhysicsWrapper = PhysicsWrapper(game_state.ball)
        self.cars: List[CarWrapper] = []
        for player in game_state.players:
            self.cars.append(CarWrapper(player_data=player))

    def blue_cars(self) -> List[CarWrapper]:
        n = len(self.cars)
        if n == 1:
            return self.cars

        return self.cars[:n // 2]

    def orange_cars(self) -> List[CarWrapper]:
        n = len(self.cars)
        if n == 1:
            return None

        return self.cars[n // 2:]

    def format_state(self) -> list:
        """
        A function to format the values stored within a StateWrapper object.
        These values are sent as a string to be applied to the game engine upon an environment reset.

        :return: String containing all state values.
        """
        # Ball: X, Y, Z, VX, VY, VZ, AVX, AVY, AVX
        # Cars: ID, X, Y, Z, VX, VY, VZ, AVX, AVY, AVZ, RX, RY, RZ, Boost

        # retrieve the ball string
        ball_state = self.ball._encode()

        # retrieve car strings
        car_states = []
        for c in self.cars:
            car_states += c._encode()

        encoded = ball_state + car_states

        return encoded


"""
A basic library for useful mathematical operations.
"""

import numpy as np


def get_dist(x, y):
    return np.subtract(x, y)


def vector_projection(vec, dest_vec, mag_squared=None):
    if mag_squared is None:
        norm = vecmag(dest_vec)
        if norm == 0:
            return dest_vec
        mag_squared = norm * norm

    if mag_squared == 0:
        return dest_vec

    dot = np.dot(vec, dest_vec)
    projection = np.multiply(np.divide(dot, mag_squared), dest_vec)
    return projection


def scalar_projection(vec, dest_vec):
    norm = vecmag(dest_vec)

    if norm == 0:
        return 0

    dot = np.dot(vec, dest_vec) / norm
    return dot


def squared_vecmag(vec):
    x = np.linalg.norm(vec)
    return x * x


def vecmag(vec):
    norm = np.linalg.norm(vec)
    return norm


def unitvec(vec):
    return np.divide(vec, vecmag(vec))


def cosine_similarity(a, b):
    return np.dot(a / np.linalg.norm(a), b / np.linalg.norm(b))


def quat_to_euler(quat):
    w, x, y, z = quat
    sinr_cosp = 2 * (w * x + y * z)
    cosr_cosp = 1 - 2 * (x * x + y * y)
    sinp = 2 * (w * y - z * x)
    siny_cosp = 2 * (w * z + x * y)
    cosy_cosp = 1 - 2 * (y * y + z * z)

    roll = np.arctan2(sinr_cosp, cosr_cosp)
    if abs(sinp) > 1:
        pitch = np.pi / 2
    else:
        pitch = np.arcsin(sinp)
    yaw = np.arctan2(siny_cosp, cosy_cosp)

    return np.array([-pitch, yaw, -roll])


# From RLUtilities
def quat_to_rot_mtx(quat: np.ndarray) -> np.ndarray:
    w = -quat[0]
    x = -quat[1]
    y = -quat[2]
    z = -quat[3]

    theta = np.zeros((3, 3))

    norm = np.dot(quat, quat)
    if norm != 0:
        s = 1.0 / norm

        # front direction
        theta[0, 0] = 1.0 - 2.0 * s * (y * y + z * z)
        theta[1, 0] = 2.0 * s * (x * y + z * w)
        theta[2, 0] = 2.0 * s * (x * z - y * w)

        # left direction
        theta[0, 1] = 2.0 * s * (x * y - z * w)
        theta[1, 1] = 1.0 - 2.0 * s * (x * x + z * z)
        theta[2, 1] = 2.0 * s * (y * z + x * w)

        # up direction
        theta[0, 2] = 2.0 * s * (x * z + y * w)
        theta[1, 2] = 2.0 * s * (y * z - x * w)
        theta[2, 2] = 1.0 - 2.0 * s * (x * x + y * y)

    return theta


def rotation_to_quaternion(m: np.ndarray) -> np.ndarray:
    trace = np.trace(m)
    q = np.zeros(4)

    if trace > 0:
        s = (trace + 1) ** 0.5
        q[0] = s * 0.5
        s = 0.5 / s
        q[1] = (m[2, 1] - m[1, 2]) * s
        q[2] = (m[0, 2] - m[2, 0]) * s
        q[3] = (m[1, 0] - m[0, 1]) * s
    else:
        if m[0, 0] >= m[1, 1] and m[0, 0] >= m[2, 2]:
            s = (1 + m[0, 0] - m[1, 1] - m[2, 2]) ** 0.5
            inv_s = 0.5 / s
            q[1] = 0.5 * s
            q[2] = (m[1, 0] + m[0, 1]) * inv_s
            q[3] = (m[2, 0] + m[0, 2]) * inv_s
            q[0] = (m[2, 1] - m[1, 2]) * inv_s
        elif m[1, 1] > m[2, 2]:
            s = (1 + m[1, 1] - m[0, 0] - m[2, 2]) ** 0.5
            inv_s = 0.5 / s
            q[1] = (m[0, 1] + m[1, 0]) * inv_s
            q[2] = 0.5 * s
            q[3] = (m[1, 2] + m[2, 1]) * inv_s
            q[0] = (m[0, 2] - m[2, 0]) * inv_s
        else:
            s = (1 + m[2, 2] - m[0, 0] - m[1, 1]) ** 0.5
            inv_s = 0.5 / s
            q[1] = (m[0, 2] + m[2, 0]) * inv_s
            q[2] = (m[1, 2] + m[2, 1]) * inv_s
            q[3] = 0.5 * s
            q[0] = (m[1, 0] - m[0, 1]) * inv_s

    # q[[0, 1, 2, 3]] = q[[3, 0, 1, 2]]

    return -q


def euler_to_rotation(pyr):
    cp, cy, cr = np.cos(pyr)
    sp, sy, sr = np.sin(pyr)

    theta = np.zeros((3, 3))

    # front
    theta[0, 0] = cp * cy
    theta[1, 0] = cp * sy
    theta[2, 0] = sp

    # left
    theta[0, 1] = cy * sp * sr - cr * sy
    theta[1, 1] = sy * sp * sr + cr * cy
    theta[2, 1] = -cp * sr

    # up
    theta[0, 2] = -cr * cy * sp - sr * sy
    theta[1, 2] = -cr * sy * sp + sr * cy
    theta[2, 2] = cp * cr

    return theta

def rand_uvec3():
    vec = np.random.rand(3) - 0.5
    return vec / np.linalg.norm(vec)

def rand_vec3(max_norm):
    return rand_uvec3() * (np.random.rand() * max_norm)



"""
The observation builder.
"""

from abc import ABC, abstractmethod
from rlgym.utils.gamestates import PlayerData, GameState
import numpy as np
from typing import Any


class ObsBuilder(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def reset(self, initial_state: GameState):
        """
        Function to be called each time the environment is reset. Note that this does not need to return anything,
        the environment will call `build_obs` automatically after reset, so the initial observation for a policy will be
        constructed in the same way as every other observation.

        :param initial_state: The initial game state of the reset environment.
        """
        raise NotImplementedError

    @abstractmethod
    def build_obs(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> Any:
        """
        Function to build observations for a policy. This is where all observations will be constructed every step and
        every reset. This function is given a player argument, and it is expected that the observation returned by this
        function will contain information from the perspective of that player. This function is called once for each
        agent automatically at every step.

        :param player: The player to build an observation for. The observation returned should be from the perspective of
        this player.

        :param state: The current state of the game.
        :param previous_action: The action taken at the previous environment step.

        :return: An observation for the player provided.
        """
        raise NotImplementedError



import numpy as np
from rlgym.envs import Match
from rlgym.utils.action_parsers import DiscreteAction
from stable_baselines3 import PPO
from stable_baselines3.common.callbacks import CheckpointCallback
from stable_baselines3.common.vec_env import VecMonitor, VecNormalize, VecCheckNan
from stable_baselines3.ppo import MlpPolicy

from rlgym.utils.obs_builders import AdvancedObs
from rlgym.utils.reward_functions.common_rewards import VelocityPlayerToBallReward
from rlgym.utils.state_setters import DefaultState
from rlgym.utils.terminal_conditions.common_conditions import TimeoutCondition, GoalScoredCondition
from rlgym_tools.sb3_utils import SB3MultipleInstanceEnv

if __name__ == '__main__':  # Required for multiprocessing
    frame_skip = 8          # Number of ticks to repeat an action
    half_life_seconds = 5   # Easier to conceptualize, after this many seconds the reward discount is 0.5

    fps = 120 / frame_skip
    gamma = np.exp(np.log(0.5) / (fps * half_life_seconds))  # Quick mafs
    print(f"fps={fps}, gamma={gamma})")


    def get_match():  # Need to use a function so that each instance can call it and produce their own objects
        return Match(
            team_size=3,  # 3v3 to get as many agents going as possible, will make results more noisy
            tick_skip=frame_skip,
            reward_function=VelocityPlayerToBallReward(),  # Simple reward since example code
            self_play=True,
            terminal_conditions=[TimeoutCondition(round(fps * 30)), GoalScoredCondition()],  # Some basic terminals
            obs_builder=AdvancedObs(),  # Not that advanced, good default
            state_setter=DefaultState(),  # Resets to kickoff position
            action_parser=DiscreteAction()  # Discrete > Continuous don't @ me
        )

    env = SB3MultipleInstanceEnv(get_match, 2)            # Start 2 instances, waiting 60 seconds between each
    env = VecCheckNan(env)                                # Optional
    env = VecMonitor(env)                                 # Recommended, logs mean reward and ep_len to Tensorboard
    env = VecNormalize(env, norm_obs=False, gamma=gamma)  # Highly recommended, normalizes rewards

    # Hyperparameters presumably better than default; inspired by original PPO paper
    model = PPO(
        MlpPolicy,
        env,
        n_epochs=32,                 # PPO calls for multiple epochs
        learning_rate=1e-5,          # Around this is fairly common for PPO
        ent_coef=0.01,               # From PPO Atari
        vf_coef=1.,                  # From PPO Atari
        gamma=gamma,                 # Gamma as calculated using half-life
        verbose=3,                   # Print out all the info as we're going
        batch_size=4096,             # Batch size as high as possible within reason
        n_steps=4096,                # Number of steps to perform before optimizing network
        tensorboard_log="out/logs",  # `tensorboard --logdir out/logs` in terminal to see graphs
        device="auto"                # Uses GPU if available
    )

    # Save model every so often
    # Divide by num_envs (number of agents) because callback only increments every time all agents have taken a step
    # This saves to specified folder with a specified name
    callback = CheckpointCallback(round(1_000_000 / env.num_envs), save_path="policy", name_prefix="rl_model")

    model.learn(100_000_000, callback=callback)

    # Now, if one wants to load a trained model from a checkpoint, use this function
    # This will contain all the attributes of the original model
    # Any attribute can be overwritten by using the custom_objects parameter,
    # which includes n_envs (number of agents), which has to be overwritten to use a different amount
    model = PPO.load("policy/rl_model_1000002_steps.zip", env, custom_objects=dict(n_envs=env.num_envs))
    env.reset()  # Important when loading models, SB3 does not do this for you
    # Use reset_num_timesteps=False to keep going with same logger/checkpoints
    model.learn(100_000_000, callback=callback, reset_num_timesteps=False)



from rlgym.utils.state_setters import StateSetter
from rlgym.utils.state_setters import StateWrapper
from rlgym.utils.math import rand_vec3
import numpy as np
from numpy import random as rand

X_MAX = 7000
Y_MAX = 9000
Z_MAX_BALL = 1850
Z_MAX_CAR = 1900
PITCH_MAX = np.pi/2
YAW_MAX = np.pi
ROLL_MAX = np.pi


class RandomState(StateSetter):

    def __init__(self, ball_rand_speed: bool = False, cars_rand_speed: bool = False, cars_on_ground: bool = True):
        """
        RandomState constructor.

        :param ball_rand_speed: Boolean indicating whether the ball will have a randomly set velocity.
        :param cars_rand_speed: Boolean indicating whether cars will have a randomly set velocity.
        :param cars_on_ground: Boolean indicating whether cars should only be placed on the ground.
        """
        super().__init__()
        self.ball_rand_speed = ball_rand_speed
        self.cars_rand_speed = cars_rand_speed
        self.cars_on_ground = cars_on_ground

    def reset(self, state_wrapper: StateWrapper):
        """
        Modifies the StateWrapper to contain random values the ball and each car.

        :param state_wrapper: StateWrapper object to be modified with desired state values.
        """
        self._reset_ball_random(state_wrapper, self.ball_rand_speed)
        self._reset_cars_random(state_wrapper, self.cars_on_ground, self.cars_rand_speed)

    def _reset_ball_random(self, state_wrapper: StateWrapper, random_speed: bool):
        """
        Function to set the ball to a random position.

        :param state_wrapper: StateWrapper object to be modified.
        :param random_speed: Boolean indicating whether to randomize velocity values.
        """
        state_wrapper.ball.set_pos(rand.random(
        ) * X_MAX - X_MAX/2, rand.random() * Y_MAX - Y_MAX/2, rand.random() * Z_MAX_BALL + 100)
        if random_speed:
            state_wrapper.ball.set_lin_vel(*rand_vec3(3000))
            state_wrapper.ball.set_ang_vel(*rand_vec3(6))

    def _reset_cars_random(self, state_wrapper: StateWrapper, on_ground: bool, random_speed: bool):
        """
        Function to set all cars to a random position.

        :param state_wrapper: StateWrapper object to be modified.
        :param on_ground: Boolean indicating whether to place cars only on the ground.
        :param random_speed: Boolean indicating whether to randomize velocity values.
        """
        for car in state_wrapper.cars:
            # set random position and rotation for all cars based on pre-determined ranges
            car.set_pos(rand.random() * X_MAX - X_MAX/2, rand.random()
                        * Y_MAX - Y_MAX/2, rand.random() * Z_MAX_CAR + 150)
            car.set_rot(rand.random() * PITCH_MAX - PITCH_MAX/2, rand.random()
                        * YAW_MAX - YAW_MAX/2, rand.random() * ROLL_MAX - ROLL_MAX/2)

            car.boost = rand.random()

            if random_speed:
                # set random linear and angular velocity based on pre-determined ranges
                car.set_lin_vel(*rand_vec3(2300))
                car.set_ang_vel(*rand_vec3(5.5))

            # 100% of cars will be set on ground if on_ground == True
            # otherwise, 50% of cars will be set on ground
            if on_ground or rand.random() < 0.5:
                # z position (up/down) is set to ground
                car.set_pos(z=17)
                # z linear velocity (vertical) set to 0
                car.set_lin_vel(z=0)
                # pitch (front of car up/down) set to 0
                # roll (side of car up/down) set to 0
                car.set_rot(pitch=0, roll=0)
                # x angular velocity (affects pitch) set to 0
                # y angular velocity (affects) roll) set to 0
                car.set_ang_vel(x=0, y=0)



import numpy as np

from rlgym.utils import TerminalCondition
from rlgym.utils.common_values import BALL_RADIUS
from rlgym.utils.gamestates import GameState
from rlgym.utils.reward_functions.common_rewards import ConstantReward
from rlgym.utils.state_setters import StateWrapper
from rlgym.utils.state_setters.random_state import RandomState, CAR_MARGIN, X_MAX, Y_MAX, ROLL_MAX, PITCH_MAX, YAW_MAX, \
    Z_MAX_BALL, Z_MAX_CAR
from rlgym_tools.suite.dribble import DribbleState, DribbleReward


class AirDribbleTerminal(TerminalCondition):
    def reset(self, initial_state: GameState):
        pass

    def is_terminal(self, current_state: GameState) -> bool:
        return not CAR_MARGIN <= current_state.players[0].car_data.position[2] <= current_state.ball.position[2]


class AirDribbleState(StateWrapper):
    def reset(self, state_wrapper: StateWrapper):
        car = state_wrapper.cars[0]
        car.set_pos(
            0,
            0,
            np.random.uniform(CAR_MARGIN, Z_MAX_CAR - 2 * BALL_RADIUS),
        )
        car.set_rot(
            pitch=np.pi / 2,
            roll=np.random.uniform(-ROLL_MAX, ROLL_MAX),
            yaw=np.random.uniform(-YAW_MAX, YAW_MAX)
        )
        r = np.random.normal(0, 10)
        theta = np.random.uniform(-np.pi, np.pi)
        eps_x, eps_y = r * np.cos(theta), r * np.sin(theta)
        state_wrapper.ball.set_pos(
            car.position[0] + eps_x,  # Displace a little bit so the ball is not balanced on the car
            car.position[1] + eps_y,
            np.random.uniform(car.position[2] + 2 * BALL_RADIUS, Z_MAX_BALL)
        )
        theta = np.random.uniform(-np.pi, np.pi)
        eps_x, eps_y = r * np.cos(theta), r * np.sin(theta)
        state_wrapper.ball.set_lin_vel(
            x=eps_x,
            y=eps_y
        )


air_dribble_config = dict(
    reward_function=DribbleReward(),
    terminal_conditions=[AirDribbleTerminal()],
    state_setter=AirDribbleState(),
    team_size=1,
    spawn_opponents=False
)



from typing import Optional, List, Union, Sequence, Type, Any

import gym
import numpy as np
from stable_baselines3.common.vec_env import VecEnv
from stable_baselines3.common.vec_env.base_vec_env import VecEnvIndices, VecEnvStepReturn, VecEnvObs

from rlgym.gym import Gym


class SB3SingleInstanceEnv(VecEnv):
    """
    Class for wrapping a single rlgym env into a VecEnv (each car is treated as its own environment).
    """

    def __init__(self, env: Gym):
        """
        :param env: the environment to wrap.
        """
        super().__init__(env._match.agents, env.observation_space, env.action_space)
        self.env = env
        self.step_result = None

    def reset(self) -> VecEnvObs:
        observations = self.env.reset()
        if len(np.shape(observations)) == 1:
            observations = [observations]
        return np.asarray(observations)

    def step_async(self, actions: np.ndarray) -> None:
        self.step_result = self.env.step(actions)

    def step_wait(self) -> VecEnvStepReturn:
        observations, rewards, done, info = self.step_result

        if type(rewards) not in (tuple, list, np.ndarray):
            rewards = [rewards]
            observations = [observations]

        if done:
            # Following what SubprocVecEnv does
            infos = [info] * len(rewards)
            for info, obs in zip(infos, observations):
                info["terminal_observation"] = obs

            observations = self.env.reset()
            if len(np.shape(observations)) == 1:
                observations = [observations]

        else:
            infos = [info] * len(rewards)
        return np.asarray(observations), np.array(rewards), np.full(len(rewards), done), infos

    def close(self) -> None:
        self.env.close()

    def seed(self, seed: Optional[int] = None) -> List[Union[None, int]]:
        return [self.env.seed(seed)] * self.num_envs

    # Now a bunch of functions that need to be overridden to work, might have to implement later
    def get_attr(self, attr_name: str, indices: VecEnvIndices = None) -> List[Any]:
        pass

    def set_attr(self, attr_name: str, value: Any, indices: VecEnvIndices = None) -> None:
        pass

    def env_method(self, method_name: str, *method_args, indices: VecEnvIndices = None, **method_kwargs) -> List[Any]:
        pass

    def env_is_wrapped(self, wrapper_class: Type[gym.Wrapper], indices: VecEnvIndices = None) -> List[bool]:
        return [False] * self.num_envs

    def get_images(self) -> Sequence[np.ndarray]:
        pass





import numpy as np
import gym.spaces
from rlgym.utils.gamestates import GameState
from rlgym.utils.action_parsers import ContinuousAction
from typing import Union, List


class DefaultAction(ContinuousAction):
    """
        Continuous Action space, that also accepts a few other input formats for QoL reasons and to remain
        compatible with older versions.
    """
    
    def __init__(self):
        super().__init__()

    def get_action_space(self) -> gym.spaces.Space:
        return super().get_action_space()

    def parse_actions(self, actions: Union[np.ndarray, List[np.ndarray], List[float]], state: GameState) -> np.ndarray:
        
        # allow other data types, this part should not be necessary but is nice to have in the default action parser.
        if type(actions) != np.ndarray:
            actions = np.asarray(actions)

        if len(actions.shape) == 1:
            actions = actions.reshape((-1, 8))
        elif len(actions.shape) > 2:
            raise ValueError('{} is not a valid action shape'.format(actions.shape))
        
        return super().parse_actions(actions, state)



import json
import os
import subprocess
import sys

import ballchasing as bc
import numpy as np
import pandas as pd

from earl_pytorch import EARL

command = r'carball.exe -i "{}" -o "{}" parquet'

ENV = os.environ.copy()
ENV["NO_COLOR"] = "1"


class CarballAnalysis:
    METADATA_FNAME = "metadata.json"
    ANALYZER_FNAME = "analyzer.json"
    BALL_FNAME = "__ball.parquet"
    GAME_FNAME = "__game.parquet"
    PLAYER_FNAME = "player_{}.parquet"

    def __init__(self, processed_folder: str):
        # print(processed_folder, self.METADATA_FNAME)
        self.metadata = json.load(open(os.path.join(processed_folder, self.METADATA_FNAME)))
        self.analyzer = json.load(open(os.path.join(processed_folder, self.ANALYZER_FNAME)))

        self.ball = pd.read_parquet(os.path.join(processed_folder, self.BALL_FNAME))
        self.game = pd.read_parquet(os.path.join(processed_folder, self.GAME_FNAME))
        self.players = {}
        for player in self.metadata["players"]:
            uid = player["unique_id"]
            player_path = os.path.join(processed_folder, self.PLAYER_FNAME.format(uid))
            if os.path.exists(player_path):
                self.players[uid] = pd.read_parquet(player_path)


def download_replays(n=1_000):
    for gamemode in bc.Playlist.RANKED:
        gm_folder = os.path.join(working_dir, "replays", gamemode)
        os.makedirs(gm_folder, exist_ok=True)
        replay_iter = api.get_replays(
            min_rank=bc.Rank.SUPERSONIC_LEGEND,
            max_rank=bc.Rank.SUPERSONIC_LEGEND,
            season=bc.Season.SEASON_5_FTP,
            count=n
        )
        for replay in replay_iter:
            if not os.path.exists(os.path.join(gm_folder, replay["id"])):
                api.download_replay(replay["id"], gm_folder)
                print(replay["id"], "downloaded")


def process_replay(replay_path, output_folder):
    folder, fn = os.path.split(replay_path)
    replay_name = fn.replace(".replay", "")
    processed_folder = os.path.join(output_folder, replay_name)
    os.makedirs(processed_folder, exist_ok=True)

    with open(os.path.join(processed_folder, "carball.o.log"), "w", encoding="utf8") as stdout_f:
        with open(os.path.join(processed_folder, "carball.e.log"), "w", encoding="utf8") as stderr_f:
            return subprocess.run(
                command.format(replay_path, processed_folder),
                stdout=stdout_f,
                stderr=stderr_f,
                env=ENV
            )


def parse_replays():
    for gamemode in bc.Playlist.RANKED:
        replay_folder = os.path.join(working_dir, "replays", gamemode)
        parsed_folder = os.path.join(working_dir, "parsed", gamemode)
        for replay in os.listdir(replay_folder):
            process_replay(os.path.join(replay_folder, replay), parsed_folder)
            print(replay, "processed")


def train_model():
    model = EARL()

    shard_size = 1_000_000
    for epoch in range(100):
        data = np.zeros((shard_size, 41, 24))

    analysis = CarballAnalysis()


def main():
    download_replays()
    parse_replays()


if __name__ == '__main__':
    working_dir = sys.argv[1]
    api = bc.Api(sys.argv[2])
    main()



import ctypes
import time
from win32con import PROCESS_ALL_ACCESS, PROCESS_SET_QUOTA, PROCESS_QUERY_INFORMATION

def page_rocket_league(rl_pid: int = -1, delay: int = 0):
    if rl_pid > 0:
        if delay > 0:
            time.sleep(delay)

        print("Paging Rocket League process id", rl_pid)

        # We require specific permissions to be able to page the process
        # See https://docs.microsoft.com/en-us/windows/win32/api/psapi/nf-psapi-emptyworkingset
        perms = PROCESS_ALL_ACCESS | PROCESS_SET_QUOTA | PROCESS_QUERY_INFORMATION

        Kernel32 = ctypes.WinDLL('kernel32.dll')
        Psapi = ctypes.WinDLL('Psapi.dll')

        try:
            # Get the process handle, trigger the page removal, and cleanup the handle
            hProcess = Kernel32.OpenProcess(perms, False, rl_pid)

            if hProcess:
                Psapi.EmptyWorkingSet(hProcess)
                Kernel32.CloseHandle(hProcess)
                return True
        except:
            # We will ignore any issues where
            pass

    return False


import random
from typing import List, Union

import numpy as np
from rlgym.utils.state_setters import StateSetter
from rlgym.utils.state_setters import StateWrapper


class ReplaySetter(StateSetter):
    def __init__(self, ndarray_or_file: Union[str, np.ndarray]):
        """
        ReplayBasedSetter constructor

        :param ndarray_or_file: A file string or a numpy ndarray of states for a single game mode.
        """
        super().__init__()

        if isinstance(ndarray_or_file, np.ndarray):
            self.states = ndarray_or_file
        elif isinstance(ndarray_or_file, str):
            self.states = np.load(ndarray_or_file)

    @classmethod
    def construct_from_replays(cls, paths_to_replays: List[str], frame_skip: int = 150):
        """
        Alternative constructor that constructs ReplayBasedSetter from replays given as paths.

        :param paths_to_replays: Paths to all the reapls
        :param frame_skip: Every frame_skip frame from the replay will be converted
        :return: Numpy array of frames
        """
        return cls(cls.convert_replays(paths_to_replays, frame_skip))

    @staticmethod
    def convert_replays(paths_to_each_replay: List[str], frame_skip: int = 150, verbose: int = 0, output_location=None):
        from rlgym_tools.replay_converter import convert_replay
        states = []
        for replay in paths_to_each_replay:
            replay_iterator = convert_replay(replay)
            remainder = random.randint(0, frame_skip - 1)  # Vary the delays slightly
            for i, value in enumerate(replay_iterator):
                if i % frame_skip == remainder:
                    game_state, _ = value

                    whole_state = []
                    ball = game_state.ball
                    ball_state = np.concatenate((ball.position, ball.linear_velocity, ball.angular_velocity))

                    whole_state.append(ball_state)
                    for player in game_state.players:
                        whole_state.append(np.concatenate((player.car_data.position,
                                                           player.car_data.euler_angles(),
                                                           player.car_data.linear_velocity,
                                                           player.car_data.angular_velocity,
                                                           np.asarray([player.boost_amount]))))

                    np_state = np.concatenate(whole_state)
                    states.append(np_state)
            if verbose > 0:
                print(replay, "done")

        states = np.asarray(states)
        if output_location is not None:
            np.save(output_location, states)
        return states

    def reset(self, state_wrapper: StateWrapper):
        """
        Modifies the StateWrapper to contain random values the ball and each car.

        :param state_wrapper: StateWrapper object to be modified with desired state values.
        """

        data = self.states[np.random.randint(0, len(self.states))]
        assert len(data) == len(state_wrapper.cars) * 13 + 9, "Data given does not match current game mode"
        self._set_ball(state_wrapper, data)
        self._set_cars(state_wrapper, data)

    def _set_cars(self, state_wrapper: StateWrapper, data: np.ndarray):
        """
        Sets the players according to the game state from replay

        :param state_wrapper: StateWrapper object to be modified with desired state values.
        :param data: Numpy array from the replay to get values from.
        """

        data = np.split(data[9:], len(state_wrapper.cars))
        for i, car in enumerate(state_wrapper.cars):
            car.set_pos(*data[i][:3])
            car.set_rot(*data[i][3:6])
            car.set_lin_vel(*data[i][6:9])
            car.set_ang_vel(*data[i][9:12])
            car.boost = data[i][12]

    def _set_ball(self, state_wrapper: StateWrapper, data: np.ndarray):
        """
        Sets the ball according to the game state from replay

        :param state_wrapper: StateWrapper object to be modified with desired state values.
        :param data: Numpy array from the replay to get values from.
        """
        state_wrapper.ball.set_pos(*data[:3])
        state_wrapper.ball.set_lin_vel(*data[3:6])
        state_wrapper.ball.set_ang_vel(*data[6:9])



"""
The Match object.
"""

from rlgym.envs.environment import Environment
from rlgym.utils.gamestates import GameState
from rlgym.utils.state_setters.wrappers.state_wrapper import StateWrapper
from rlgym.utils import common_values
import gym.spaces
import numpy as np
from typing import List, Union, Any


class Match(Environment):
    def __init__(self,
                 reward_function,
                 terminal_conditions,
                 obs_builder,
                 action_parser,
                 state_setter,
                 team_size=1,
                 tick_skip=8,
                 game_speed=100,
                 spawn_opponents=False,
                 self_play=False):
        super().__init__()

        self._game_speed = game_speed
        self._team_size = team_size
        self._self_play = self_play
        self._spawn_opponents = spawn_opponents or self_play
        self._tick_skip = tick_skip
        self._reward_fn = reward_function
        self._terminal_conditions = terminal_conditions
        self._obs_builder = obs_builder
        self._action_parser = action_parser
        self._state_setter = state_setter

        if type(terminal_conditions) not in (tuple, list):
            self._terminal_conditions = [terminal_conditions, ]

        self.agents = self._team_size * 2 if self._self_play else self._team_size

        self.observation_space = None
        self._auto_detect_obs_space()
        self.action_space = self._action_parser.get_action_space()

        self._prev_actions = np.zeros((self.agents, 8), dtype=float)
        self._spectator_ids = None

        self.last_touch = None
        self._initial_score = 0

    def episode_reset(self, initial_state: GameState):
        self._spectator_ids = [p.car_id for p in initial_state.players]
        self._prev_actions.fill(0)
        for condition in self._terminal_conditions:
            condition.reset(initial_state)
        self._reward_fn.reset(initial_state)
        self._obs_builder.reset(initial_state)
        self.last_touch = None
        self._initial_score = initial_state.blue_score - initial_state.orange_score

    def build_observations(self, state) -> Union[Any, List]:
        observations = []
        for i in range(len(state.players)):
            player = state.players[i]
            if player.team_num == common_values.ORANGE_TEAM and not self._self_play:
                continue
            else:
                obs = self._obs_builder.build_obs(player, state, self._prev_actions[i])

            observations.append(obs)

        if state.last_touch is None:
            state.last_touch = self.last_touch
        else:
            self.last_touch = state.last_touch

        if len(observations) == 1:
            return observations[0]

        return observations

    def get_rewards(self, state, done) -> Union[float, List]:
        rewards = []

        for i in range(len(state.players)):
            player = state.players[i]

            if player.team_num == common_values.ORANGE_TEAM and not self._self_play:
                continue

            if done:
                reward = self._reward_fn.get_final_reward(player, state, self._prev_actions[i])
            else:
                reward = self._reward_fn.get_reward(player, state, self._prev_actions[i])

            rewards.append(reward)

        if len(rewards) == 1:
            return rewards[0]

        return rewards

    def is_done(self, state):
        for condition in self._terminal_conditions:
            if condition.is_terminal(state):
                return True
        return False

    def get_result(self, state: GameState):
        current_score = state.blue_score - state.orange_score
        return current_score - self._initial_score

    def parse_state(self, state_str: List[float]) -> GameState:
        state = GameState(state_str)
        return state

    def parse_actions(self, actions: Any, state: GameState) -> np.ndarray:
        # Prevent people from accidentally modifying numpy arrays inside the ActionParser
        if isinstance(actions, np.ndarray):
            actions = np.copy(actions)
        return self._action_parser.parse_actions(actions, state)

    def format_actions(self, actions: np.ndarray):
        self._prev_actions[:] = actions[:]

        acts = []
        for i in range(len(actions)):
            acts.append(float(self._spectator_ids[i]))
            for act in actions[i]:
                acts.append(float(act))

        return acts

    def get_reset_state(self) -> list:
        new_state = StateWrapper(blue_count=self._team_size,
                                 orange_count=self._team_size if self._spawn_opponents == True else 0)
        self._state_setter.reset(new_state)
        return new_state.format_state()

    def get_reset_state(self) -> str:
        new_state = StateWrapper(blue_count=self._team_size,
                                 orange_count=self._team_size if self._spawn_opponents == True else 0)
        self._state_setter.reset(new_state)
        return new_state.format_state()

    def get_config(self):
        return [self._team_size,
                1 if self._self_play else 0,
                1 if self._spawn_opponents else 0,
                self._tick_skip,
                self._game_speed]

    def _auto_detect_obs_space(self):
        from rlgym.utils.gamestates.player_data import PlayerData

        num_cars = self._team_size*2 if self._spawn_opponents else self._team_size
        empty_player_packets = []
        for i in range(num_cars):
            player_packet = PlayerData()
            player_packet.car_id = i
            empty_player_packets.append(player_packet)

        empty_game_state = GameState()
        prev_inputs = np.zeros(common_values.NUM_ACTIONS)

        empty_game_state.players = empty_player_packets

        obs_shape = np.shape(self._obs_builder.build_obs(empty_player_packets[0], empty_game_state, prev_inputs))

        self.observation_space = gym.spaces.Box(-np.inf, np.inf, shape=obs_shape)


import time
from collections import deque
from typing import List

import gym
import numpy as np
import torch as th
from stable_baselines3 import PPO
from stable_baselines3.common import utils
from stable_baselines3.common.callbacks import BaseCallback
from stable_baselines3.common.type_aliases import GymEnv, MaybeCallback
from stable_baselines3.common.utils import obs_as_tensor, safe_mean
from stable_baselines3.common.vec_env import VecEnv
from typing import Dict, Optional, Union

# This is to allow sequential multi_learn calls
globs = {
"LAST_ALL_OBS" : None,
"LAST_MODEL_MAP" : None,
"OBS_SIZE" : None
} # type: Dict[str, Optional[Union[int, list]]]


# This function is heavily based off the collect_rollouts() method of the sb3 OnPolicyAlgorithm
def multi_collect_rollouts(
        env: VecEnv, models: List[PPO], model_map: list, all_last_obs: list,  n_rollout_steps: int,
        obs_size: int, all_callbacks: List[BaseCallback], learning_mask: List[bool]):
    n_steps = 0
    all_last_episode_restarts = [models[model_map[num]]._last_episode_starts for num in range(len(model_map))]
    for model in models:
        model.rollout_buffer.reset()

    for callback in all_callbacks: callback.on_rollout_start()

    models_length = len(models)
    map_length = len(model_map)
    while n_steps < n_rollout_steps:
        # create indexes to replace later
        all_actions = [0 for _ in range(map_length)]
        all_values = [0 for _ in range(map_length)]
        all_log_probs = [0 for _ in range(map_length)]
        all_clipped_actions = [0 for _ in range(map_length)]

        # disgusting dict and list comprehension to put the same model obs together
        per_model_obs = { model_index:
            np.array([
                all_last_obs[obs_index] for obs_index in range(map_length) if model_map[obs_index] == model_index
            ])
            for model_index in range(models_length) if model_index in model_map
        }

        for model_index in range(models_length):
            # get the actions from the policy
            if model_index in model_map:
                with th.no_grad():
                    obs_tensor = obs_as_tensor(per_model_obs[model_index], models[model_index].device)
                    actions, values, log_probs = models[model_index].policy.forward(obs_tensor)
                actions = actions.cpu().numpy()
                clipped_actions = actions #[0] # it is inside an extra layer for some reason, so take it out
                if isinstance(models[model_index], gym.spaces.Box):
                    clipped_actions = np.clip(
                        actions,
                        models[model_index].action_space.low,
                        models[model_index].action_space.high
                    )

                next_index_start = 0
                # put everything back in terms of the model map
                for i in range(model_map.count(model_index)):
                    next_index = model_map.index(model_index, next_index_start)
                    next_index_start = next_index + 1
                    all_clipped_actions[next_index] = clipped_actions[i]
                    all_actions[next_index] = actions[i]
                    all_values[next_index] = values[i]
                    all_log_probs[next_index] = log_probs[i]
        # flatten the actions, then step the env
        flat_clipped_actions = np.array(all_clipped_actions)
        flat_new_obs, flat_rewards, flat_dones, flat_infos = env.step(flat_clipped_actions)
        # split up the returns from the step
        infos_length = len(flat_infos) // map_length
        all_infos = [flat_infos[x*infos_length:(x+1)*infos_length] for x in range(map_length)]
        all_rewards = [flat_rewards[x] for x in range(map_length)]

        # increment num_timesteps
        for obs_index in range(map_length):
            models[model_map[obs_index]].num_timesteps += 1

        # allow the callbacks to run
        for callback in all_callbacks: callback.update_locals(locals())
        if any(callback.on_step() is False for callback in all_callbacks):
            return False, all_last_obs

        # update the info buffer for each model
        for model_index in range(models_length):
            models[model_index]._update_info_buffer(
                [all_infos[num][0] for num in range(map_length) if model_map[num] == model_index]
            ) # this should put the needed infos for each model in
        n_steps += 1

        # reshape for models with discrete action spaces
        for obs_index in range(map_length):
            if isinstance(models[model_map[obs_index]].action_space, gym.spaces.Discrete):
                all_actions[obs_index] = all_actions[obs_index].reshape(-1,1)

        # add data to the rollout buffers
        for model_index in range(models_length):
            if learning_mask[model_index] and model_index in model_map: # skip learing where not necessary
                models[model_index].rollout_buffer.add( # disgusting list comprehension to send all the info to the buffer
                    np.asarray([all_last_obs[num][0] for num in range(len(model_map)) if model_map[num] == model_index]),
                    np.asarray([all_actions[num] for num in range(len(model_map)) if model_map[num] == model_index]),
                    np.asarray([all_rewards[num] for num in range(len(model_map)) if model_map[num] == model_index]),
                    np.asarray([all_last_episode_restarts[num] for num in range(len(model_map)) if model_map[num] == model_index]),
                    th.tensor([all_values[num] for num in range(len(model_map)) if model_map[num] == model_index]),
                    th.tensor([all_log_probs[num] for num in range(len(model_map)) if model_map[num] == model_index])
                )

        # shuffle variables for next iteration
        new_obs_len = len(flat_new_obs) // map_length
        all_last_obs = [flat_new_obs[obs_index * new_obs_len:(obs_index + 1) * new_obs_len] for obs_index in range(map_length)]
        all_last_episode_restarts = flat_dones

    all_last_values, all_last_dones = [], []
    for obs_index in range(len(model_map)):
        with th.no_grad():
            # compute value for the last timestamp
            # the og code uses new_obs where I have last_obs, so I hope this still works since they should hold the same value
            obs_tensor = obs_as_tensor(all_last_obs[obs_index], models[model_map[obs_index]].device)
            _, values, _ = models[model_map[obs_index]].policy.forward(obs_tensor)
            all_last_values.append(values)

    # compute the returns and advantage for each model
    for model_index in range(len(models)):
        if model_index in model_map:
            models[model_index].rollout_buffer.compute_returns_and_advantage(
                last_values=th.tensor([all_last_values[num] for num in range(len(model_map)) if model_map[num] == model_index]),
                dones=np.asarray([all_last_episode_restarts[num] for num in range(len(model_map)) if model_map[num] == model_index])
            )

    for callback in all_callbacks: callback.on_rollout_end()
    return True, all_last_obs

# This function is heavily based off the learn() method of the sb3 OnPolicyAlgorithm
def multi_learn(
        models: List[PPO],
        total_timesteps: int,
        env,
        num_players: int,
        learning_mask: Optional[List[bool]] = None,
        model_map: Optional[list] = None,
        callbacks: List[MaybeCallback] = None,
        log_interval: int = 1,
        eval_env: Optional[GymEnv] = None,
        eval_freq: int = -1,
        n_eval_episodes: int = 5,
        tb_log_name: str = "MultiPPO",
        eval_log_path: Optional[str] = None,
        reset_num_timesteps: bool = True,
):
    model_map = model_map or [n % len(models) for n in range(num_players)]
    learning_mask = learning_mask or [True for _ in range(len(models))]

    # make sure everything lines up
    if not len(models) == len(callbacks) == len(learning_mask):
        raise ValueError("Length of models, callbacks, and learning_mask must all be equal.")

    iteration = 0
    # this for loop is essentially the setup method, done for each model
    if globs['OBS_SIZE'] is None:
        globs['OBS_SIZE'] = len(env.reset()) // len(model_map)  # calculate the length of the each observation
    all_total_timesteps = []
    for model_index in range(len(models)):
        models[model_index].start_time = time.time()
        if models[model_index].ep_info_buffer is None or reset_num_timesteps:
            models[model_index].ep_info_buffer = deque(maxlen=100)
            models[model_index].ep_success_buffer = deque(maxlen=100)

        if models[model_index].action_noise is not None:
            models[model_index].action_noise.reset()

        if reset_num_timesteps:
            models[model_index].num_timesteps = 0
            models[model_index]._episode_num = 0
            all_total_timesteps.append(total_timesteps)
            models[model_index]._total_timesteps = total_timesteps
        else:
            # make sure training timestamps are ahead of internal counter
            all_total_timesteps.append(total_timesteps + models[model_index].num_timesteps)
            models[model_index]._total_timesteps = total_timesteps + models[model_index].num_timesteps

        # leaving out the environment reset that normally happens here, since that will be done for all at once

        if eval_env is not None and models[model_index].seed is not None:
            eval_env.seed(models[model_index].seed)

        eval_env = models[model_index]._get_eval_env(eval_env)

        # Configure logger's outputs if no logger was passed
        if not models[model_index]._custom_logger:
            models[model_index]._logger = utils.configure_logger(
                models[model_index].verbose,
                models[model_index].tensorboard_log,
                tb_log_name + f'_model{model_index}',
                reset_num_timesteps)

        callbacks[model_index] = models[model_index]._init_callback(
            callbacks[model_index], eval_env, eval_freq, n_eval_episodes, log_path=None)


    for callback in callbacks: callback.on_training_start(locals(), globals())
    if globs["LAST_ALL_OBS"] is None:
        flat_last_obs = env.reset()
        globs["LAST_ALL_OBS"] = [flat_last_obs[x*globs['OBS_SIZE']:(x+1)*globs['OBS_SIZE']] for x in range(num_players)]


    # make sure the n_envs is correct for the models
    for model_index in range(len(models)):
        models[model_index].n_envs = model_map.count(model_index)
        models[model_index].rollout_buffer.n_envs = model_map.count(model_index)

    # I assume the correct thing here is to check each model separately for the while condition
    while all([models[i].num_timesteps < all_total_timesteps[i] for i in range(len(models))]):
        continue_training, globs["LAST_ALL_OBS"] = multi_collect_rollouts(
            env, models, model_map, globs["LAST_ALL_OBS"], min(model.n_steps for model in models), globs["OBS_SIZE"], callbacks, learning_mask
        )

        if continue_training is False:
            break

        iteration += 1
        for model_index in range(len(models)):
            if model_index in model_map:
                models[model_index]._update_current_progress_remaining(models[model_index].num_timesteps, total_timesteps)

        # output to the logger
        for model_index in range(len(models)):
            if log_interval is not None and iteration % log_interval == 0 and learning_mask[model_index] and model_index in model_map:
                fps = int(models[model_index].num_timesteps / (time.time() - models[model_index].start_time))
                models[model_index].logger.record("time/iterations", iteration * model_map.count(model_index), exclude="tensorboard")
                if len(models[model_index].ep_info_buffer) > 0 and len(models[model_index].ep_info_buffer[0]) > 0:
                    models[model_index].logger.record("rollout/ep_rew_mean", safe_mean([ep_info["r"] for ep_info in models[model_index].ep_info_buffer]))
                    models[model_index].logger.record("rollout/ep_len_mean", safe_mean([ep_info["l"] for ep_info in models[model_index].ep_info_buffer]))
                models[model_index].logger.record("time/fps", fps)
                models[model_index].logger.record("time/time_elapsed", int(time.time() - models[model_index].start_time), exclude="tensorboard")
                models[model_index].logger.record("time/total_timesteps", models[model_index].num_timesteps, exclude="tensorboard")
                models[model_index].logger.dump(step=models[model_index].num_timesteps)

        for model_index in range(len(models)):
            if learning_mask[model_index] and model_index in model_map: models[model_index].train()


    for callback in callbacks: callback.on_training_end()


    return models


from typing import Optional

import torch
from torch import nn as nn

from ..util.constants import DEFAULT_FEATURES


class EPRL(nn.Module):  # Experimental alternative to EARL, P stands for pooling-based
    def __init__(self, n_dims: int = 512, n_hidden: int = 4, n_features: Optional[int] = None):
        super().__init__()
        self.n_dims = n_dims
        self.n_features = n_features or len(DEFAULT_FEATURES)

        self.initial_fully_connected = nn.Sequential(*(
                [nn.Linear(self.n_features, n_dims), nn.ReLU()]
                + sum(([nn.Linear(n_dims, n_dims), nn.ReLU()] for _ in range(n_hidden)), start=[])
        ))

        self.max_pool = nn.AdaptiveMaxPool1d(1)

        self.final_fc = nn.Sequential(*(
                [nn.Linear(4 * n_dims, n_dims), nn.ReLU()]
                + sum(([nn.Linear(n_dims, n_dims), nn.ReLU()] for _ in range(n_hidden)), start=[])
        ))

    def forward(self, main_player: torch.Tensor, other_players: torch.Tensor,
                balls: torch.Tensor, boosts: torch.Tensor):
        main_player_emb = self.initial_fully_connected(main_player)
        other_players_emb = self.initial_fully_connected(other_players)
        balls_emb = self.initial_fully_connected(balls)
        boosts_emb = self.initial_fully_connected(boosts)

        main_player_emb = self.max_pool(main_player_emb.swapdims(1, 2)).squeeze(2)
        other_players_emb = self.max_pool(other_players_emb.swapdims(1, 2)).squeeze(2)
        balls_emb = self.max_pool(balls_emb.swapdims(1, 2)).squeeze(2)
        boosts_emb = self.max_pool(boosts_emb.swapdims(1, 2)).squeeze(2)

        emb = torch.cat((main_player_emb, other_players_emb, balls_emb, boosts_emb), dim=-1)

        emb = self.final_fc(emb)

        return emb



from abc import abstractmethod

import numpy as np

from rlgym.utils import RewardFunction
from rlgym.utils.common_values import BLUE_TEAM, ORANGE_TEAM
from rlgym.utils.gamestates import PlayerData, GameState


class ConditionalRewardFunction(RewardFunction):
    def __init__(self, reward_func: RewardFunction):
        super().__init__()
        self.reward_func = reward_func

    @abstractmethod
    def condition(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> bool:
        raise NotImplementedError

    def reset(self, initial_state: GameState):
        pass

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        if self.condition(player, state, previous_action):
            return self.reward_func.get_reward(player, state, previous_action)
        return 0

    def get_final_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        if self.condition(player, state, previous_action):
            return self.reward_func.get_final_reward(player, state, previous_action)
        return 0


class RewardIfClosestToBall(ConditionalRewardFunction):
    def __init__(self, reward_func: RewardFunction, team_only=True):
        super().__init__(reward_func)
        self.team_only = team_only

    def condition(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> bool:
        dist = np.linalg.norm(player.car_data.position - state.ball.position)
        for player2 in state.players:
            if not self.team_only or player2.team_num == player.team_num:
                dist2 = np.linalg.norm(player2.car_data.position - state.ball.position)
                if dist2 < dist:
                    return False
        return True


class RewardIfTouchedLast(ConditionalRewardFunction):
    def condition(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> bool:
        return state.last_touch == player.car_id


class RewardIfBehindBall(ConditionalRewardFunction):
    def condition(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> bool:
        return player.team_num == BLUE_TEAM and player.car_data.position[1] < state.ball.position[1] \
               or player.team_num == ORANGE_TEAM and player.car_data.position[1] > state.ball.position[1]


from rlgym.utils.reward_functions import RewardFunction
from rlgym.utils import math
from rlgym.utils.gamestates import GameState, PlayerData
import numpy as np


# THIS REWARD IS FOR DEMONSTRATION PURPOSES ONLY
class DefaultReward(RewardFunction):
    def __init__(self):
        super().__init__()
        self.last_touch = None

    def reset(self, initial_state: GameState):
        self.last_touch = initial_state.last_touch

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        return - math.vecmag(player.car_data.angular_velocity) / 100

    def get_final_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        return self.get_reward(player, state, previous_action)



"""
Data classes to permit the manipulation of environment variables.
"""

from rlgym.utils.gamestates.player_data import PlayerData
from rlgym.utils.gamestates.physics_object import PhysicsObject
from rlgym.utils.gamestates.game_state import GameState
from typing import List
import numpy as np

BLUE_ID1 = 1
ORANGE_ID1 = 5


class StateWrapper(object):

    def __init__(self, blue_count: int = 0, orange_count: int = 0, game_state=None):
        """
        StateWrapper constructor. Under most circumstances, users should not expect to instantiate their own StateWrapper objects.

        :param blue_count: Integer indicating the amount of players on the blue team.
        :param orange_count: Integer indicating The amount of players on the orange team.
        :param game_state: GameState object for values to be copied from.

        NOTE: blue_count and orange_count will be ignored if a GameState object is passed.
        """
        if game_state is None:
            self.ball: PhysicsWrapper = PhysicsWrapper()
            self.cars: List[CarWrapper] = []
            for i in range(blue_count):
                self.cars.append(CarWrapper(0, BLUE_ID1 + i))
            for i in range(orange_count):
                self.cars.append(CarWrapper(1, ORANGE_ID1 + i))
        else:
            self._read_from_gamestate(game_state)

    def _read_from_gamestate(self, game_state: GameState):
        """
        A function to modify the StateWrapper with values read in from a GameState object.
        """
        self.ball: PhysicsWrapper = PhysicsWrapper(game_state.ball)
        self.cars: List[CarWrapper] = []
        for player in game_state.players:
            self.cars.append(CarWrapper(player_data=player))

    def format_state(self) -> str:
        """
        A function to format the values stored within a StateWrapper object. 
        These values are sent as a string to be applied to the game engine upon an environment reset.

        :return: String containing all state values.
        """
        # Ball: X, Y, Z, VX, VY, VZ, AVX, AVY, AVX
        # Cars: ID, X, Y, Z, VX, VY, VZ, AVX, AVY, AVZ, RX, RY, RZ, Boost

        # retrieve the ball string
        ball_str = self.ball._encode()

        # retrieve car strings
        car_str_list = [c._encode() for c in self.cars]

        return f'{ball_str} {" ".join(car_str_list)}'


class PhysicsWrapper(object):

    def __init__(self, phys_obj: PhysicsObject = None):
        """
        PhysicsWrapper constructor. Under most circumstances, users should not expect to instantiate their own PhysicsWrapper objects.

        :param phys_obj: PhysicsObject object from which values will be read.
        """
        if phys_obj is None:
            self.position: np.ndarray = np.zeros(3)
            self.linear_velocity: np.ndarray = np.zeros(3)
            self.angular_velocity: np.ndarray = np.zeros(3)
        else:
            self._read_from_physics_object(phys_obj)

    def _read_from_physics_object(self, phys_obj: PhysicsObject):
        """
        A function to modify PhysicsWrapper values from values in a PhysicsObject object.
        """
        self.position = phys_obj.position
        self.linear_velocity = phys_obj.linear_velocity
        self.angular_velocity = phys_obj.angular_velocity

    def set_pos(self, x: float = None, y: float = None, z: float = None):
        """
        Sets position.

        :param x: Float indicating x position value.
        :param y: Float indicating y position value.
        :param z: Float indicating z position value.
        """
        if x is not None:
            self.position[0] = x
        if y is not None:
            self.position[1] = y
        if z is not None:
            self.position[2] = z

    def set_lin_vel(self, x: float = None, y: float = None, z: float = None):
        """
        Sets linear velocity.

        :param x: Float indicating x velocity value.
        :param y: Float indicating y velocity value.
        :param z: Float indicating z velocity value.
        """
        if x is not None:
            self.linear_velocity[0] = x
        if y is not None:
            self.linear_velocity[1] = y
        if z is not None:
            self.linear_velocity[2] = z

    def set_ang_vel(self, x: float = None, y: float = None, z: float = None):
        """
        Sets angular velocity.

        :param x: Float indicating x angular velocity value.
        :param y: Float indicating y angular velocity value.
        :param z: Float indicating z angular velocity value.
        """
        if x is not None:
            self.angular_velocity[0] = x
        if y is not None:
            self.angular_velocity[1] = y
        if z is not None:
            self.angular_velocity[2] = z

    def _encode(self) -> str:
        """
        Function called by a StateWrapper to produce a state string.

        :return: String containing value data.
        """
        state_arr = np.concatenate((self.position,
                                    self.linear_velocity, self.angular_velocity), dtype=str)
        return " ".join(state_arr)


class CarWrapper(PhysicsWrapper):

    def __init__(self, team_num: int = -1, id: int = -1, player_data: PlayerData = None):
        """
        CarWrapper constructor. Under most circumstances, users should not expect to instantiate their own CarWrapper objects.

        :param team_num: Integer indicating 0 for blue and 1 for orange.
        :param id: Integer indicating the spectator ID assigned to the car.
        :param player_data: PlayerData object for values to be copied from.

        NOTE: team_num and id will be ignored if a PlayerData object is passed.
        """
        if player_data is None:
            super().__init__()
            self.rotation: np.ndarray = np.zeros(3)
            self.team_num: int = team_num
            self.id: int = id
            self.boost: float = 0
        else:
            super().__init__(phys_obj=player_data.car_data)
            self._read_from_player_data(player_data)

    def _read_from_player_data(self, player_data: PlayerData):
        """
        A function to modify CarWrapper values from values in a PlayerData object.
        """
        self.rotation = player_data.car_data.euler_angles()
        self.team_num = player_data.team_num
        self.id = player_data.car_id
        self.boost = player_data.boost_amount

    def set_rot(self, pitch: float = None, yaw: float = None, roll: float = None):
        """
        Sets rotation in terms of euler angles.
        """
        if pitch is not None:
            self.rotation[0] = pitch
        if yaw is not None:
            self.rotation[1] = yaw
        if roll is not None:
            self.rotation[2] = roll

    def _encode(self) -> str:
        """
        Function called by a StateWrapper to produce a state string.

        :return: String containing value data.
        """
        state_arr = np.concatenate((self.position, self.linear_velocity,
                                    self.angular_velocity, self.rotation), dtype=str)
        return f'{self.id} {" ".join(state_arr)} {self.boost}'



from rlgym.utils.obs_builders import ObsBuilder
from rlgym.utils import common_values, math
import numpy as np


class RhobotObs(ObsBuilder):
    def __init__(self):
        super().__init__()

    def reset(self, initial_state):
        pass

    def build_obs(self, player, state, prev_action) -> np.ndarray:
        if prev_action is None:
            print("!ATTEMPTED TO BUILD RHOBOT OBS WITH NO PREV ACTIONS ARGUMENT!")
            raise AssertionError

        players = state.players
        if player.team_num == common_values.ORANGE_TEAM:
            player_car = player.inverted_car_data
            ball = state.inverted_ball
        else:
            player_car = player.car_data
            ball = state.ball

        ob = []
        ob.append(prev_action)
        ob.append([int(player.has_flip),
                   player.boost_amount,
                   int(player.on_ground)])

        ob.append(player_car.position)
        ob.append(player_car.euler_angles())
        ob.append(np.sin(player_car.euler_angles()))
        ob.append(np.cos(player_car.euler_angles()))
        yaw = player_car.yaw()
        angle_between_bot_and_target = np.arctan2(ball.position[1] - player_car.position[1],
                                                  ball.position[0] - player_car.position[0])
        
        angle_front_to_target = angle_between_bot_and_target - yaw
        ob.append([angle_front_to_target])
        ob.append(player_car.linear_velocity)
        ob.append(player_car.angular_velocity)

        ob.append(ball.position)
        ob.append(ball.linear_velocity)
        ob.append(ball.angular_velocity)

        ob.append(common_values.ORANGE_GOAL_CENTER)
        ob.append(common_values.BLUE_GOAL_CENTER)

        pb_dist = math.vecmag(math.get_dist(player_car.position, ball.position))
        ob.append([pb_dist])

        # Since we invert the car and ball data when the agent is in the Orange team
        # the "Orange Goal" is always the enemy goal
        pg_dist = math.vecmag(math.get_dist(player_car.position, common_values.ORANGE_GOAL_CENTER))
        ob.append([pg_dist])

        pog_dist = math.vecmag(math.get_dist(player_car.position, common_values.BLUE_GOAL_CENTER))
        ob.append([pog_dist])

        for other in players:
            if other.car_id == player.car_id:
                continue

            if player.team_num == common_values.ORANGE_TEAM:
                car_data = other.inverted_car_data
            else:
                car_data = other.car_data

            ob.append(car_data.position)
            ob.append(car_data.euler_angles())
            ob.append(np.sin(car_data.euler_angles()))
            ob.append(np.cos(car_data.euler_angles()))
            ob.append(car_data.linear_velocity)
            ob.append(car_data.angular_velocity)

            pc_dist = math.vecmag(math.get_dist(player_car.position, car_data.position))
            ob.append([pc_dist])

        return np.concatenate(ob)



from typing import Any, Optional, Tuple, overload, Union

import numpy as np
from rlgym.utils.reward_functions import RewardFunction
from rlgym.utils.gamestates import GameState, PlayerData


class CombinedReward(RewardFunction):
    """
    A reward composed of multiple rewards.
    """

    def __init__(
            self,
            reward_functions: Tuple[RewardFunction, ...],
            reward_weights: Optional[Tuple[float, ...]] = None
    ):
        """
        Creates the combined reward using multiple rewards, and a potential set
        of weights for each reward.

        :param reward_functions: Each individual reward function.
        :param reward_weights: The weights for each reward.
        """
        super().__init__()

        self.reward_functions = reward_functions
        self.reward_weights = reward_weights or np.ones_like(reward_functions)

        if len(self.reward_functions) != len(self.reward_weights):
            raise ValueError(
                ("Reward functions list length ({0}) and reward weights " \
                 "length ({1}) must be equal").format(
                    len(self.reward_functions), len(self.reward_weights)
                )
            )

    @classmethod
    def from_zipped(cls, *rewards_and_weights: Union[RewardFunction, Tuple[RewardFunction, float]]) -> "CombinedReward":
        """
        Alternate constructor which takes any number of either rewards, or (reward, weight) tuples.

        :param rewards_and_weights: a sequence of RewardFunction or (RewardFunction, weight) tuples
        """
        rewards = []
        weights = []
        for value in rewards_and_weights:
            if isinstance(value, tuple):
                r, w = value
            else:
                r, w = value, 1.
            rewards.append(r)
            weights.append(w)
        return cls(tuple(rewards), tuple(weights))

    def reset(self, initial_state: GameState) -> None:
        """
        Resets underlying reward functions.

        :param initial_state: The initial state of the reset environment.
        """
        for func in self.reward_functions:
            func.reset(initial_state)

    def get_reward(
            self,
            player: PlayerData,
            state: GameState,
            previous_action: np.ndarray
    ) -> float:
        """
        Returns the reward for a player on the terminal state.

        :param player: Player to compute the reward for.
        :param state: The current state of the game.
        :param previous_action: The action taken at the previous environment step.

        :return: The combined rewards for the player on the state.
        """
        rewards = [
            func.get_reward(player, state, previous_action)
            for func in self.reward_functions
        ]

        return float(np.dot(self.reward_weights, rewards))

    def get_final_reward(
            self,
            player: PlayerData,
            state: GameState,
            previous_action: np.ndarray
    ) -> float:
        """
        Returns the reward for a player on the terminal state.

        :param player: Player to compute the reward for.
        :param state: The current state of the game.
        :param previous_action: The action taken at the previous environment step.

        :return: The combined rewards for the player on the state.
        """
        rewards = [
            func.get_final_reward(player, state, previous_action)
            for func in self.reward_functions
        ]

        return float(np.dot(self.reward_weights, rewards))



# From https://github.com/RLBot/RLBot/blob/master/src/main/python/rlbot/version.py
# Store the version here so:
# 1) we don't load dependencies by storing it in __init__.py
# 2) we can import it in setup.py for the same reason
# 3) we can import it into your module module
# https://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package

__version__ = '1.6.1'

release_notes = {
    '1.6.1': {
        """
        - Fixed angular velocities in replay augmenter
        """
    },
    '1.6.0': {
        """
        - Added GoaliePracticeState setter (Soren)
        - Added ReplaySetter (Carrot)
        - Added AugmentSetter (NeverCast)
        - Fixed an error in replay converter
        """
    },
    '1.5.3': {
        """
        - Yet another fix for GameCondition
        """
    },
    '1.5.2': {
        """
        - Another fix for GameCondition
        """
    },
    '1.5.1': {
        """
        - Fix for GameCondition
        """
    },
    '1.5.0': {
        """
        - Add SB3CombinedLogReward (LuminousLamp)
        - Add SB3InstantaneousFPSCallback (LuminousLamp)
        - Add KickoffLikeSetter (Carrot)
        - Add GameCondition (Rolv)
        """
    },
    '1.4.1': {
        """
        - Remove SB3MultiDiscreteWrapper
        - Update SB3 examples to include action parsers
        """
    },
    '1.4.0': {
        """
        - Add KBM action parser
        """
    },
    '1.3.0': 
    """
    - Add KickoffReward (Impossibum)
    - Add SequentialReward (Impossibum)
    - Better individual reward logging for SB3 (LuminousLamp)
    - Add launch_preference as parameter to SB3MultipleInstanceEnv
    """,
    '1.2.0':
    """
    - Added multi-model support for SB3
    - Added weighted sample setter
    - Added general stacker
    - Various bug fixes
    """,
    '1.1.4':
    """
    - Bug fixes for AdvancedStacker
    """,
    '1.1.3':
    """
    - Fix RLGym 1.0 incompatibility
    """,
    '1.1.2':
    """
    - Fix version import (was absolute, now relative)
    """,
    '1.1.1':
    """
    - Fixed an invalid Python 3.7 import
    - AnnealRewards now has initial_count parameter
    """,
    '1.1.0': """
    - Added functionality in SB3MultipleInstanceEnv to take multiple match objects (one for each instance).
    - Added AnnealRewards for transitioning between reward functions.
    - Added granularity option to SB3MultiDiscreteWrapper.
    - Added negative slope parameter to DiffReward (negative values are multiplied by this).
    """,
    '1.0.0': """
    - Added replay to rlgym GameState converter
    - Moved SB3 environments from rlgym (now called SB3SingleInstanceEnv and SB3MultipleInstanceEnv) and fixed some bugs
    - Added SB3MultiDiscreteWrapper, SB3DistributeRewardsWrapper and SB3LogReward 
    - Added extra reward functions (DiffReward, DistributeRewards and MultiplyRewards)
    - Added RLLibEnv
    - Added working example code for SB3 and RLlib
    """
}


def get_current_release_notes():
    if __version__ in release_notes:
        return release_notes[__version__]
    return ''


def print_current_release_notes():
    print(f"Version {__version__}")
    print(get_current_release_notes())
    print("")



import platform
from dataclasses import dataclass
import subprocess
import webbrowser
from pathlib import Path
from typing import Optional
from rlgym.gamelaunch.epic_launch import launch_with_epic_simple, launch_with_epic_login_trick
import os

# Copied from https://github.com/RLBot/RLBot/blob/master/src/main/python/rlbot/setup_manager.py

class ROCKET_LEAGUE_PROCESS_INFO:
    GAMEID = 252950
    PROGRAM_NAME = 'RocketLeague.exe'
    PROGRAM = 'RocketLeague.exe'
    REQUIRED_ARGS = {r'-pipe'}

    @staticmethod
    def get_ideal_args(pipe_id):
        # We are specifying RLBot_PacketSendRate=240, which will override people's TARLBot.ini settings.
        # We believe there is no downside to 240. See https://github.com/RLBot/RLBot/wiki/Tick-Rate
        return ['-pipe', f'{pipe_id}', '-nomovie']


class LaunchPreference:
    STEAM = 'steam'
    EPIC = 'epic'
    EPIC_LOGIN_TRICK = EPIC + '_login_trick'


def run_injector():
    print("Executing injector...")
    cur_dir = os.path.dirname(os.path.realpath(__file__))
    injector_command = os.path.join(cur_dir, os.pardir, "plugin", "RLMultiInjector.exe")
    subprocess.Popen([injector_command])


def launch_rocket_league(pipe_id, launch_preference: str = LaunchPreference.EPIC) -> Optional[subprocess.Popen]:
    """
    Launches Rocket League but does not connect to it.
    """
    ideal_args = ROCKET_LEAGUE_PROCESS_INFO.get_ideal_args(pipe_id)

    if not launch_preference.startswith((LaunchPreference.EPIC, LaunchPreference.STEAM)):
        if os.path.isfile(launch_preference):
            return subprocess.Popen([launch_preference] + ideal_args)
        else:
            print("path_to_rl doesn't point to RocketLeague.exe")

    if launch_preference.startswith(LaunchPreference.EPIC):
        if launch_preference == LaunchPreference.EPIC_LOGIN_TRICK:
            proc = launch_with_epic_login_trick(ideal_args)
            if proc is not None:
                print('Launched with Epic login trick')
                return proc
            else:
                print("Epic login trick seems to have failed, falling back to simple Epic launch.")
        # Fall back to simple if the tricks failed or we opted out of tricks.
        game_process = launch_with_epic_simple(ideal_args)
        if game_process:
            print('Launched Epic version')
            return game_process

    # Try launch via Steam.
    steam_exe_path = try_get_steam_executable_path()
    if steam_exe_path:  # Note: This Python 3.8 feature would be useful here https://www.python.org/dev/peps/pep-0572/#abstract
        exe_and_args = [
                           str(steam_exe_path),
                           '-applaunch',
                           str(ROCKET_LEAGUE_PROCESS_INFO.GAMEID),
                       ] + ideal_args
        # print(f'Launching Rocket League with: {exe_and_args}')
        _ = subprocess.Popen(exe_and_args)  # This is deliberately an orphan process.
        print('Launched Steam version')
        return

    print(f'Launching Rocket League using Steam-only fall-back launch method with args: {ideal_args}')
    print("You should see a confirmation pop-up, if you don't see it then click on Steam! "
                     'https://gfycat.com/AngryQuickFinnishspitz')
    args_string = '%20'.join(ideal_args)

    # Try launch via terminal (Linux)
    if platform.system() == 'Linux':
        linux_args = [
            'steam',
            f'steam://rungameid/{ROCKET_LEAGUE_PROCESS_INFO.GAMEID}//{args_string}'
        ]

        try:
            _ = subprocess.Popen(linux_args)
            print('Launched Steam Linux version')
            return
        except OSError:
            print('Could not launch Steam executable on Linux.')

    try:
        print("Launching rocket league via steam browser URL as a last resort...")
        webbrowser.open(f'steam://rungameid/{ROCKET_LEAGUE_PROCESS_INFO.GAMEID}//{args_string}')
    except webbrowser.Error:
        print(f'Unable to launch Rocket League. Please launch Rocket League manually using the -pipe {pipe_id} option to continue.')


def try_get_steam_executable_path() -> Optional[Path]:
    """
    Tries to find the path of the Steam executable.
    Has platform specific code.
    """

    try:
        from winreg import OpenKey, HKEY_CURRENT_USER, ConnectRegistry, QueryValueEx, REG_SZ
    except ImportError as e:
        return  # TODO: Linux support.

    try:
        key = OpenKey(ConnectRegistry(None, HKEY_CURRENT_USER), r'Software\Valve\Steam')
        val, val_type = QueryValueEx(key, 'SteamExe')
    except FileNotFoundError:
        return
    if val_type != REG_SZ:
        return
    return Path(val)


import torch.distributions
from all.agents import PPO, Agent
from all.approximation import FeatureNetwork, VNetwork, QNetwork
from all.bodies import Body
from all.experiments import run_experiment
from all.policies import SoftmaxPolicy, DeterministicPolicy, SoftDeterministicPolicy, GaussianPolicy
from all.policies.deterministic import DeterministicPolicyNetwork
from all.presets import Preset, PresetBuilder
from gym.spaces import Box, MultiDiscrete, Discrete
from torch import nn
from torch.nn import Sequential
from torch.optim import Adam

import rlgym
from rlgym_tools.autonomous_utils.autonomous_env import autonomous_env


class RLPreset(Preset):
    def __init__(self, env, name, device, **hyperparameters):
        super().__init__(name, device, hyperparameters)
        self.feature_model = nn.Linear(72, 128).to(device)
        self.value_model = nn.Linear(128, 1).to(device)
        self.policy_model = nn.Linear(128, 2 * 8).to(device)

    def agent(self, writer=None, train_steps=float('inf')):
        lr = self.hyperparameters.pop("lr")
        eps = self.hyperparameters.pop("eps")
        feature_optimizer = Adam(self.feature_model.parameters(), lr=lr,
                                 eps=eps)
        value_optimizer = Adam(self.value_model.parameters(), lr=lr,
                               eps=eps)
        policy_optimizer = Adam(self.policy_model.parameters(), lr=lr,
                                eps=eps)

        # approximators
        f = FeatureNetwork(self.feature_model, feature_optimizer, writer=writer)
        v = VNetwork(self.value_model, value_optimizer, writer=writer)
        p = GaussianPolicy(self.policy_model, policy_optimizer, space=Box(-1, 1, (8,)), writer=writer)

        return PPO(f, v, p, **self.hyperparameters, n_envs=1)

    def test_agent(self):
        pass


class RLBody(Body):
    def process_state(self, state):
        return

if __name__ == '__main__':
    env = autonomous_env(rlgym.make(team_size=1, self_play=True))

    agents = PresetBuilder("rlgym", default_hyperparameters={
        "discount_factor": 0.99,
        "lr": 2e-4,
        "eps": 1.5e-4,
        "minibatches": 32,
    }, constructor=RLPreset, env=env, device="cuda")

    # agents = [
    #     PresetBuilder("rlgym", None, RLPreset, env=env)
    #     for n in range(2)
    # ]
    run_experiment(agents, env, frames=1000)
